@@ -6,7 +6,7 @@ from datetime import datetime
 import gzip
 import requests # type: ignore
 import ijson # type: ignore
-from dwh_oppfolging.transforms.functions import dict_to_string, string_to_sha256_hash, string_to_naive_norwegian_datetime
+from dwh_oppfolging.transforms.functions import json_to_string, string_to_sha256_hash, string_to_naive_norwegian_datetime
 from dwh_oppfolging.apis.brreg_api_v1_types import PagedRequestForEmbeddedList, Update, PagedRequestError, UnpagedRequestForEmbeddedList
 
 
@@ -52,8 +52,8 @@ def _build_unit_record(update: Update, fact: dict, download_date: datetime):
     record["endringstype"] = update.change
     record["oppdatert_tid_kilde"] = update.last_modified
     record["api_versjon"] = API_VERSION
-    record["data"] = dict_to_string(fact)
-    record["sha256_hash"] = string_to_sha256_hash(record["data"] + record["oppdatert_tid_kilde"].isoformat())
+    record["data"] = json_to_string(fact)
+    record["sha256_hash"] = string_to_sha256_hash(record["data"])
     record["lastet_dato"] = download_date
     record["kildesystem"] = API_NAME
     return record
