@@ -146,7 +146,7 @@ def update_table_from_sql(
     bind_today_to_etl_date: bool = True,
     etl_date_bind_name: str = "etl_date",
     enable_etl_logging: bool = True,
-) -> None:
+) -> tuple[int, int]:
     """
     basic update of table using provided sql
     if bind_today_to_etl_date is set then today() is bound to variable :etl_date_bind_name
@@ -166,13 +166,15 @@ def update_table_from_sql(
     logging.info(f"updated {rows_updated} existing rows")
     if enable_etl_logging:
         log_etl(cur, schema, table, today, rows_inserted, rows_updated)
+    return rows_inserted, rows_updated
 
 
 def build_insert_sql_string(
     schema: str,
     table: str,
     cols: list[str],
-    filter_columns: list[str] | None = None
+    unique_columns: list[str] | None = None,
+    additional_where_clauses: list[str] | None = None,
 ) -> str:
     """
     returns a formattable sql insert, optionally with filter columns,
@@ -180,49 +182,142 @@ def build_insert_sql_string(
     with the same column values already exist.
     target table columns are formatted with targ_cols,
     bind columns (values to insert) are formatted with bind_cols
+    NB: additional where clauses must not use the 'where' keyword
     >>> build_insert_sql_string('a', 'b', ['x', 'y'], ['x', 'y'])
     'insert into a.b targ (targ.x, targ.y) select :x, :y from dual src where not exists (select null from a.b t where t.x = :x and t.y = :y)'
+    >>> build_insert_sql_string('a', 'b', ['x', 'y'], ['x', 'y'], ["2 = 3", "5 = 4"])
+    'insert into a.b targ (targ.x, targ.y) select :x, :y from dual src where not exists (select null from a.b t where t.x = :x and t.y = :y) and 2 = 3 and 5 = 4'
+    >>> build_insert_sql_string('a', 'b', ['x', 'y'], None, ["2 = 3", "5 = 4"])
+    'insert into a.b targ (targ.x, targ.y) select :x, :y from dual src where 2 = 3 and 5 = 4'
     """
     targ_cols = ", targ.".join(cols)
     bind_cols = ", :".join(cols)
     sql = (
         f"insert into {schema}.{table} targ (targ.{targ_cols}) select :{bind_cols} from dual src"
     )
-    if filter_columns is not None and len(filter_columns) > 0:
+    where_set = False
+    if unique_columns is not None and len(unique_columns) > 0:
         sql += (
             f" where not exists (select null from {schema}.{table} t where "
-            + " and ".join(f"t.{col} = :{col}" for col in filter_columns)
+            + " and ".join(f"t.{col} = :{col}" for col in unique_columns)
             + ")"
         )
+        where_set = True
+    if additional_where_clauses is not None and len(additional_where_clauses) > 0:
+        if not where_set:
+            sql += " where "
+        else:
+            sql += " and "
+        sql += " and ".join(clause for clause in additional_where_clauses)
     return sql
 
 
-def insert_to_table_batchwise(
+def _insert_to_table_gen(
     cur: Cursor,
     schema: str,
     table: str,
-    batch_generator: Generator[list[Row], None, None],
-    filter_columns: list[str] | None = None,
+    data: Generator[list[Row] | Row, None, None] | list[Row] | Row,
+    unique_columns: list[str] | None = None,
+    additional_where_clauses: list[str] | None = None,
+    enable_etl_logging: bool = False,
+    continue_on_db_errors: bool = False
+):
+    # check if data is Row
+    if isinstance(data, dict):
+        data = [data]
+
+    # insert data
+    insert_sql = ""
+    rows_inserted = 0
+    for item in data:
+        if not isinstance(item, list):
+            item = [item]
+        elif len(item) == 0:
+            continue
+        if not insert_sql:
+            cols = [*(item[0])]
+            insert_sql = build_insert_sql_string(schema, table, cols, unique_columns, additional_where_clauses)
+        cur.executemany(insert_sql, item, batcherrors=continue_on_db_errors)
+        batcherrors = cur.getbatcherrors() or []
+        rows_inserted += cur.rowcount
+        yield (cur.rowcount, batcherrors)
+    if enable_etl_logging:
+        log_etl(cur, schema, table, datetime.today(), rows_inserted)
+
+
+def insert_to_table(
+    cur: Cursor,
+    schema: str,
+    table: str,
+    data: Generator[list[Row] | Row, None, None] | list[Row] | Row,
+    unique_columns: list[str] | None = None,
+    additional_where_clauses: list[str] | None = None,
     enable_etl_logging: bool = True,
 ):
     """
-    inserts into table batch-wise from generator
-    it is assumed that the column names are the same in all batches,
-    and always appear, and in the same order.
-    if enable etl logging is set, log_etl is called for the target table
-    with etl_date today()
+    Inserts data into table. No commits are made.
+    Data can be a row, list of rows, or a generator of either.
+    Returns number of rows inserted.
+
+    `unique_columns`: if provided, this combination of columns must be unique for 
+    each row to be inserted, or it is skipped. Default: None
+
+    `enable_etl_logging`: if set, metadata will be inserted into the etl logging table
+    at the end of data insertion. Default: True
     """
     rows_inserted = 0
-    is_sql_formatted = False
-    insert_sql = ""
-    for batch in batch_generator:
-        if len(batch) == 0:
-            continue
-        if not is_sql_formatted:
-            is_sql_formatted = True
-            cols = [*(batch[0])]
-            insert_sql = build_insert_sql_string(schema, table, cols, filter_columns)
-        cur.executemany(insert_sql, batch)
-        rows_inserted += cur.rowcount
-    if enable_etl_logging:
-        log_etl(cur, schema, table, datetime.today())
+    for info in _insert_to_table_gen(
+        cur,
+        schema,
+        table, data,
+        unique_columns,
+        additional_where_clauses,
+        enable_etl_logging
+    ):
+        rows_inserted += info[0]
+    return rows_inserted
+
+
+def create_table_insert_generator(
+    cur: Cursor,
+    schema: str,
+    table: str,
+    data: Generator[list[Row] | Row, None, None] | list[Row] | Row,
+    unique_columns: list[str] | None = None,
+    additional_where_clauses: list[str] | None = None,
+    enable_etl_logging: bool = True,
+    continue_on_db_errors: bool = False,
+):
+    """
+    Creates a generator that inserts data into a table. No commits are made.
+
+    Data can be a row, list of rows, or a generator of either.
+
+    The generator yields tuples of insert information 
+    (rows inserted, [error]). Errors are only returned if `continue_on_db_errors` is set,
+    otherwise the list is empty. The error objects are tuples of
+    (batch_index, errcode, message).
+
+    `unique_columns`: if provided, this combination of columns must be unique for 
+    each row to be inserted, or it is skipped. Default: None
+
+    `enable_etl_logging`: if set, metadata will be inserted into the etl logging table
+    at the end of data insertion. Default: True
+
+    `continue_on_db_errors`: if set, then ORA errors are yielded at each batch and insertion
+    is allowed to continue, rather than throwing an exception. Otherwise, only array dml
+    counts are returned. Default: False
+    NOTE: DPY errors are still thrown, for example when trying to insert a string into
+    a number column.
+    """
+    for info in _insert_to_table_gen(
+        cur,
+        schema,
+        table,
+        data,
+        unique_columns,
+        additional_where_clauses,
+        enable_etl_logging,
+        continue_on_db_errors
+    ):
+        yield info
