@@ -3,4 +3,4 @@
 from datetime import datetime
 
 
-Row = dict[str, str | int | float | bytes | datetime]
+Row = dict[str, str | int | float | bytes | datetime | None]
